@@ -66,6 +66,7 @@ class Assign():
         self.atom_marker = {}
         self.bonds = {}
         self.ar_bonds = {}
+        self.am_bonds = {}
         self.bond_marker = {}
 
     def Atom_Judge(self, atom, string):
@@ -120,6 +121,10 @@ class Assign():
             else:
                 self.atom_marker[atom2][marker] = 1
 
+    def Determine_Equal_Atoms(self):
+        from .RDKit_tools import Find_Equal_Atoms
+        return Find_Equal_Atoms(self)
+
     def Determine_Bond_Order(self):
         raise NotImplementedError
 
@@ -262,7 +267,7 @@ class Assign():
                     find_type = True
                     break
 
-            assert find_type
+            assert find_type, "No atom type found for assignment %s of atom #%d"%(self.name, i)
         # print(self.atom_types)
 
     def To_ResidueType(self, name, charge=None):
@@ -311,7 +316,6 @@ class Assign():
         towrite = towrite = "REMARK   Generated By Xponge (Assignment)\n"
         count = {}
         for i in range(self.atom_numbers):
-            assert self.atom_types[i] != None
             if self.names[i]:
                 atom_name = self.names[i]
             elif self.atoms[i] in count.keys():
@@ -328,16 +332,18 @@ class Assign():
                                                                                  resname, " ", 1, self.coordinate[i][0],
                                                                                  self.coordinate[i][1],
                                                                                  self.coordinate[i][2], " ", atom)
+
         for i in range(self.atom_numbers):
             bonded_atoms = list(self.bonds[i].keys())
             bonded_atoms.sort()
             bonded_atoms = [bonded_atoms[i:i + 4] for i in range(0, len(bonded_atoms), 4)]
             if len(bonded_atoms) > 0:
                 for atoms in bonded_atoms:
-                    towrite += "CONECT %4d" % i
+                    towrite += "CONECT %4d" % (i + 1)
                     for atom in atoms:
-                        towrite += " %4d" % atom
+                        towrite += " %4d" % (atom + 1)
                     towrite += "\n"
+
         f = open(filename, "w")
         f.write(towrite)
         f.close()
@@ -349,6 +355,8 @@ class Assign():
                 if i < j:
                     if  i in self.ar_bonds.keys() and j in self.ar_bonds[i]:
                         bonds.append("%6d %6d ar\n" % (i + 1, j + 1))
+                    elif i in self.am_bonds.keys() and j in self.am_bonds[i]:
+                        bonds.append("%6d %6d am\n" % (i + 1, j + 1))
                     else:
                         bonds.append("%6d %6d %1d\n" % (i + 1, j + 1, order))
         bonds.sort(key=lambda x: list(map(int, x.split()[:2])))
@@ -382,6 +390,51 @@ class Assign():
         f.close()
 
 
+def Guess_Element_From_Mass(mass):
+    elements = ["X",  "H",  "He", "Li", "Be", "B",  "C",  "N",  "O",  "F",  "Ne",
+    "Na", "Mg", "Al", "Si", "P" , "S",  "Cl", "Ar", "K",  "Ca", "Sc",
+    "Ti", "V",  "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga", "Ge", 
+    "As", "Se", "Br", "Kr", "Rb", "Sr", "Y",  "Zr", "Nb", "Mo", "Tc",
+    "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te", "I",  "Xe",
+    "Cs", "Ba", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb",
+    "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta", "W",  "Re", "Os",
+    "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr",
+    "Ra", "Ac", "Th", "Pa", "U",  "Np", "Pu", "Am", "Cm", "Bk", "Cf",
+    "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt",
+    "Ds", "Rg"]
+    masses =  [ 0.00000, 1.00794, 4.00260, 6.941, 9.012182, 10.811,  
+    12.0107, 14.0067, 15.9994, 18.9984032, 20.1797, 
+    22.989770, 24.3050, 26.981538, 28.0855, 30.973761,
+    32.065, 35.453, 39.948, 39.0983, 40.078, 44.955910,
+    47.867, 50.9415, 51.9961, 54.938049, 55.845, 58.9332,
+    58.6934, 63.546, 65.409, 69.723, 72.64, 74.92160, 
+    78.96, 79.904, 83.798, 85.4678, 87.62, 88.90585, 
+    91.224, 92.90638, 95.94, 98.0, 101.07, 102.90550,
+    106.42, 107.8682, 112.411, 114.818, 118.710, 121.760, 
+    127.60, 126.90447, 131.293, 132.90545, 137.327, 
+    138.9055, 140.116, 140.90765, 144.24, 145.0, 150.36,
+    151.964, 157.25, 158.92534, 162.500, 164.93032, 
+    167.259, 168.93421, 173.04, 174.967, 178.49, 180.9479,
+    183.84, 186.207, 190.23, 192.217, 195.078, 196.96655, 
+    200.59, 204.3833, 207.2, 208.98038, 209.0, 210.0, 222.0, 
+    223.0, 226.0, 227.0, 232.0381, 231.03588, 238.02891,
+    237.0, 244.0, 243.0, 247.0, 247.0, 251.0, 252.0, 257.0,
+    258.0, 259.0, 262.0, 261.0, 262.0, 266.0, 264.0, 269.0,
+    268.0, 271.0, 272.0 ]
+    if mass > 0.0 and mass < 3.8:
+        index = 1;
+    elif mass > 207.85 and mass < 208.99: 
+        index = 83;
+    elif mass > 56.50 and mass < 58.8133:
+        index = 27;
+    else:
+        index = 0;
+        for j in range(0,111):
+            if abs(mass - masses[j]) < 0.65:
+                index = j
+                break
+    return elements[index]
+
 def Get_Assignment_From_PubChem(parameter, keyword):
     cs = pcp.get_compounds(parameter, keyword, record_type='3d')
     if len(cs) == 0:
@@ -399,11 +452,14 @@ def Get_Assignment_From_PubChem(parameter, keyword):
         raise NotImplementedError
 
 
-def Get_Assignment_From_PDB(filename):
+def Get_Assignment_From_PDB(filename, determine_bond_order = True):
     assign = Assign()
+    index_atom_map = {}
     with open(filename) as f:
         for line in f:
             if line.startswith("ATOM") or line.startswith("HETATM"):
+                index = int(line[6:11])
+                index_atom_map[index] = assign.atom_numbers
                 atom_name = line[12:16].strip()
                 element = line[76:78].strip()
                 x = float(line[30:38])
@@ -411,13 +467,25 @@ def Get_Assignment_From_PDB(filename):
                 z = float(line[46:54])
                 assign.Add_Atom(element, x, y, z, atom_name)
             if line.startswith("CONECT"):
-                atom = int(line[6:11]) - 1
+                atom = int(line[6:11])
                 for bonded_atom in line[11:].split():
-                    assign.Add_Bond(atom, int(bonded_atom) - 1)
-    assign.Determine_Bond_Order()
-    assign.Determine_Ring_And_Bond_Type()
+                    assign.Add_Bond(index_atom_map[atom], index_atom_map[int(bonded_atom)])
+    if determine_bond_order:
+        assign.Determine_Bond_Order()
+        assign.Determine_Ring_And_Bond_Type()
     return assign
 
+def Get_Assignment_From_ResidueType(restype):
+    assign = Assign()
+    for atom in restype.atoms:
+        assign.Add_Atom(Guess_Element_From_Mass(atom.mass), atom.x, atom.y, atom.z, atom.name)
+    for atom in restype.atoms:
+        i = restype._atom2index[atom]
+        for atomb in restype.connectivity[atom]:
+            j = restype._atom2index[atomb]
+            if i < j:
+                assign.Add_Bond(i, j)
+    return assign
 
 def Get_Assignment_From_Mol2(filename):
     with open(filename) as f:
@@ -458,6 +526,18 @@ def Get_Assignment_From_Mol2(filename):
                         assign.ar_bonds[atom2] = [atom1]
                     else:
                         assign.ar_bonds[atom2].append(atom1)
+                elif words[3] == "am":
+                    atom1 = int(words[1]) - 1
+                    atom2 = int(words[2]) - 1
+                    assign.Add_Bond(atom1, atom2, 1)
+                    if atom1 not in assign.am_bonds.keys():
+                        assign.am_bonds[atom1] = [atom2]
+                    else:
+                        assign.am_bonds[atom1].append(atom2)
+                    if atom2 not in assign.am_bonds.keys():
+                        assign.am_bonds[atom2] = [atom1]
+                    else:
+                        assign.am_bonds[atom2].append(atom1)
                 else:
                     raise NotImplementedError("No implemented method to process bond #%s type %s"%(words[0], words[3]))
     ar_bonds_atoms = list(assign.ar_bonds.keys())
@@ -491,7 +571,32 @@ def Get_Assignment_From_Mol2(filename):
     assign.Determine_Ring_And_Bond_Type()
     return assign
 
+def Get_Assignment_From_SDF(filename):
+    with open(filename) as f:
+        assign = Assign(f.readline().strip())
+        f.readline()
+        f.readline()
+        line = f.readline()
+        atom_numbers = int(line[:3].strip())
+        bond_numbers = int(line[3:6].strip())
+        for i in range(atom_numbers):
+            line = f.readline()
+            element = line[31:33].strip()
+            x = float(line[:10])
+            y = float(line[10:20])
+            z = float(line[20:30])
+            assign.Add_Atom(element, x, y, z)
+        for i in range(bond_numbers):
+            line = f.readline()
+            atom1 = int(line[:3]) - 1
+            atom2 = int(line[3:6]) - 1
+            bond_order = int(line[6:9])
+            assign.Add_Bond(atom1, atom2, bond_order)
+        
+    assign.Determine_Ring_And_Bond_Type()
+    return assign
 
 sys.modules["__main__"].__dict__["Get_Assignment_From_PubChem"] = Get_Assignment_From_PubChem
 sys.modules["__main__"].__dict__["Get_Assignment_From_PDB"] = Get_Assignment_From_PDB
 sys.modules["__main__"].__dict__["Get_Assignment_From_Mol2"] = Get_Assignment_From_Mol2
+sys.modules["__main__"].__dict__["Get_Assignment_From_ResidueType"] = Get_Assignment_From_ResidueType
